"""JSON encoding functions using EAK-Rules."""

import datetime
import decimal

from peak.rules import abstract
from prioritized_methods import prioritized_when, prioritized_around

from simplejson import JSONEncoder


# Global options

descent_bases = True


# Specific encoding functions

@abstract()
def jsonify(obj):
    """Generic function for converting objects to JSON.

    Specific functions should return a string or an object that can be
    serialized with JSON, i.e., it is made up of only lists, dictionaries
    (with string keys), and strings, ints, and floats.

    """
    raise NotImplementedError

# This is for easier usage and backward compatibility:
jsonify.when = prioritized_when.__get__(jsonify)
jsonify.around = prioritized_around.__get__(jsonify)


@jsonify.when("isinstance(obj, (datetime.date, datetime.datetime))", prio=-1)
def jsonify_datetime(obj):
    """JSONify datetime and date objects."""
    return str(obj)

@jsonify.when("isinstance(obj, decimal.Decimal)", prio=-1)
def jsonify_decimal(obj):
    """JSONify decimal objects."""
    return float(obj)

@jsonify.when("hasattr(obj, '__json__')", prio=-1)
def jsonify_explicit(obj):
    """JSONify objects with explicit JSONification method."""
    return obj.__json__()


# SQLObject support

try:
    from sqlobject import SQLObject

    def _sqlobject_attrs(obj):
        """Get all attributes of an SQLObject."""
        sm = obj.__class__.sqlmeta
        try:
            while sm is not None:
                # we need to exclude the ID-keys, as for some reason
                # this won't work for subclassed items
                for key in sm.columns:
                    if key[-2:] != 'ID':
                        yield key
                sm = descent_bases and sm.__base__ or None
        except AttributeError: # happens if we descent to <type object>
            pass

    def is_sqlobject(obj):
        return (isinstance(obj, SQLObject)
            and hasattr(obj.__class__, 'sqlmeta'))

    @jsonify.when("is_sqlobject(obj) and not hasattr(obj, '__json__')", prio=-1)
    def jsonify_sqlobject(obj):
        """JSONify SQLObjects."""
        result = {'id': obj.id}
        for name in _sqlobject_attrs(obj):
            result[name] = getattr(obj, name)
        return result

    try:
        SelectResultsClass = SQLObject.SelectResultsClass
    except AttributeError:
        pass
    else:

        @jsonify.when("isinstance(obj, SelectResultsClass)", prio=-1)
        def jsonify_select_results(obj):
            """JSONify SQLObject.SelectResults."""
            return list(obj)

except ImportError:
    pass


# SQLAlchemy support

try:
    import sqlalchemy

    try:
        import sqlalchemy.ext.selectresults
        from sqlalchemy.util import OrderedProperties
    except ImportError: # SQLAlchemy >= 0.5

        def is_saobject(obj):
            return hasattr(obj, '_sa_class_manager')

        @jsonify.when("is_saobject(obj) and not hasattr(obj, '__json__')",
                      prio=-1)
        def jsonify_saobject(obj):
            """JSONify SQLAlchemy objects."""
            props = {}
            for key in obj.__dict__:
                if not key.startswith('_sa_'):
                    props[key] = getattr(obj, key)
            return props

    else: # SQLAlchemy < 0.5

        def is_saobject(obj):
            return (hasattr(obj, 'c')
                and isinstance(obj.c, OrderedProperties))

        @jsonify.when("is_saobject(obj) and not hasattr(obj, '__json__')",
                      prio=-1)
        def jsonify_saobject(obj):
            """JSONify SQLAlchemy objects."""
            props = {}
            for key in obj.c.keys():
                props[key] = getattr(obj, key)
            return props

        try:
            from sqlalchemy.orm.attributes import InstrumentedList
        except ImportError: # SQLAlchemy >= 0.4
            pass # normal lists are used here

        else: # SQLAlchemy < 0.4

            @jsonify.when("isinstance(obj, InstrumentedList)", prio=-1)
            def jsonify_instrumented_list(obj):
                """JSONify SQLAlchemy instrumented lists."""
                return list(obj)

except ImportError:
    pass


# JSON Encoder class

class GenericJSON(JSONEncoder):

    def __init__(self, **opts):
        opt = opts.pop('descent_bases', None)
        if opt is not None:
            global descent_bases
            descent_bases = opt
        super(GenericJSON, self).__init__(**opts)

    def default(self, obj):
        return jsonify(obj)

_instance = GenericJSON()


# General encoding functions

def encode(obj):
    """Return a JSON string representation of a Python object."""
    return _instance.encode(obj)

def encode_iter(obj):
    """Encode object, yielding each string representation as available."""
    return _instance.iterencode(obj)



from turbojson import jsonify

try:
    try:
        import sqlite3
    except:
        import pysqlite2
    from sqlalchemy import MetaData, Table, Column, ForeignKey, \
        Integer, String
    from sqlalchemy.orm import create_session, mapper, relation

    metadata = MetaData('sqlite:///:memory:')
    test1 = Table('test1', metadata,
            Column('id', Integer, primary_key=True),
            Column('val', String(8)))

    test2 = Table('test2', metadata,
            Column('id', Integer, primary_key=True),
            Column('test1id', Integer, ForeignKey('test1.id')),
            Column('val', String(8)))

    test3 = Table('test3', metadata,
            Column('id', Integer, primary_key=True),
            Column('val', String(8)))
    test4 = Table('test4', metadata,
            Column('id', Integer, primary_key=True),
            Column('val', String(8)))
    metadata.create_all()

    class Test2(object):
        pass
    mapper(Test2, test2)

    class Test1(object):
        pass
    mapper(Test1, test1, properties={'test2s':relation(Test2)})

    class Test3(object):
        def __json__(self):
            return {'id': self.id, 'val': self.val, 'customized': True}

    mapper(Test3, test3)

    class Test4(object):
        pass
    mapper(Test4, test4)

    test1.insert().execute({'id':1, 'val':'bob'})
    test2.insert().execute({'id':1, 'test1id':1, 'val':'fred'})
    test2.insert().execute({'id':2, 'test1id':1, 'val':'alice'})
    test3.insert().execute({'id':1, 'val':'bob'})
    test4.insert().execute({'id':1, 'val':'alberto'})

except ImportError:
    from warnings import warn
    warn('SQLAlchemy or PySqlite not installed - cannot run these tests.')

else:

    def test_saobj():
        s = create_session()
        t = s.query(Test1).get(1)
        encoded = jsonify.encode(t)
        assert encoded == '{"id": 1, "val": "bob"}'

    def test_salist():
        s = create_session()
        t = s.query(Test1).get(1)
        encoded = jsonify.encode(t.test2s)
        assert encoded == '[{"test1id": 1, "id": 1, "val": "fred"},' \
            ' {"test1id": 1, "id": 2, "val": "alice"}]'

    def test_explicit_saobj():
        s = create_session()
        t = s.query(Test3).get(1)
        encoded = jsonify.encode(t)
        assert encoded == '{"id": 1, "val": "bob", "customized": true}'

    def test_priority_override():
        s = create_session()
        t = s.query(Test4).get(1)
        encoded = jsonify.encode(t)
        assert encoded == '{"id": 1, "val": "alberto"}'

        @jsonify.jsonify.when((Test4,))
        def jsonify_test4(obj):
            return {'val':obj.val}

        encoded = jsonify.encode(t)
        assert encoded == '{"val": "alberto"}'
        



from turbojson import jsonify

try:
    try:
        import sqlite3
    except:
        import pysqlite2
    from sqlobject import sqlhub, connectionForURI, SQLObject, StringCol
    from sqlobject.inheritance import InheritableSQLObject

    sqlhub.processConnection = connectionForURI('sqlite:/:memory:')

    class Person(SQLObject):
        fname = StringCol()
        mi = StringCol(length=1, default=None)
        lname = StringCol()
    Person.createTable()

    class ExplicitPerson(SQLObject):
        fname = StringCol()
        def __json__(self):
            return {'customized': True}
    ExplicitPerson.createTable()

    class A(InheritableSQLObject):
        foo = StringCol()
    A.createTable()

    class B(A):
        bar = StringCol()
    B.createTable()

    class C(A):
        baz = StringCol()
        def __json__(self):
            return {'customized': True}
    C.createTable()

except ImportError:
    from warnings import warn
    warn('SQLObject or PySqlite not installed - cannot run these tests.')

else:

    def test_soobj():
        p = Person(fname="Peter", mi="P", lname="Pasulke")
        pj = jsonify.jsonify(p)
        assert pj == {'fname': "Peter", 'mi': "P", 'lname': "Pasulke", 'id': 1}
        b = B(foo="foo", bar="bar")
        bj = jsonify.jsonify(b)
        assert bj == {'foo': "foo", 'bar': "bar", 'id': b.id, 'childName': None}

    def test_customized_soobj():
        ep = ExplicitPerson(fname="Peter")
        epj = jsonify.jsonify(ep)
        assert epj == {'customized': True}
        c = C(foo="foo", baz="baz")
        cj = jsonify.jsonify(c)
        assert cj == {'customized': True}

    def test_so_select_result():
        p = Person(fname="Willy", mi="P", lname="Millowitsch")
        sr = Person.select(Person.q.fname=="Willy")
        srj = jsonify.jsonify(sr)
        assert srj == [p]

    def test_descent_bases():
        b = B(foo="foo", bar="bar")
        bj = jsonify.jsonify(b)
        assert 'foo' in bj and 'bar' in bj
        no_descent_encoder = jsonify.GenericJSON(descent_bases=False)
        bj = no_descent_encoder.encode(b)
        assert 'foo' not in bj and 'bar' in bj



"""Template support for JSON"""

from turbojson import jsonify


class JsonSupport(object):

    encoding = 'utf-8'

    def __init__(self, extra_vars_func=None, options=None):
        opts = {}
        for option in options:
            if not option.startswith('json.'):
                continue
            opt = option[5:]
            if opt == 'encoding':
                self.encoding = options[option] or self.encoding
            else:
                if opt == 'assume_encoding':
                    opt = 'encoding'
                opts[opt] = options[option]
        self.encoder = jsonify.GenericJSON(**opts)

    def render(self, info, format=None, fragment=False, template=None):
        """Renders data in the desired format.

        @param info: the data itself
        @type info: dict
        @param format: not used
        @type format: string
        @param fragment: not used
        @type fragment: bool
        @param template: not used
        @type template: string

        """
        for item in info.keys():
            if item.startswith("tg_") and item != "tg_flash":
                del info[item]
        output = self.encoder.encode(info)
        if isinstance(output, unicode):
            output = output.encode(self.encoding)
        return output



# -*- coding: UTF-8 -*-

from setuptools import setup, find_packages

setup(
    name = 'TurboJson',
    version = '1.2.1',
    description = 'Python template plugin that supports JSON',
    author = 'Elvelind Grandin',
    author_email = 'elvelind+turbogears@gmail.com',
    maintainer = 'TurboGears project',
    maintainer_email =  'turbogears@googlegroups.com',
    url = 'http://docs.turbogears.org/TurboJson',
    download_url = 'http://pypi.python.org/pypi/TurboJson',
    license = 'MIT',
    keywords = [
        'python.templating.engines',
        'turbogears'
    ],
    install_requires = [
        'PEAK-Rules >= 0.5a1.dev-r2555',
        'simplejson >= 1.9.1',
        'prioritized_methods >= 0.2'
    ],
    zip_safe = True,
    packages = find_packages(),
    classifiers = [
        'Development Status :: 4 - Beta',
        'Framework :: TurboGears',
        'Environment :: Web Environment :: Buffet',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    entry_points = """\
    [python.templating.engines]
    json = turbojson.jsonsupport:JsonSupport
    """,
    test_suite = 'nose.collector'
)



from turbojson import jsonify


class Foo(object):
    def __init__(self, bar):
        self.bar = bar

class Bar(object):
    def __init__(self, bar):
        self.bar = bar
    def __json__(self):
        return 'bar-%s' % self.bar

class Baz(object):
    pass


def jsonify_foo(obj):
    return "foo-%s" % obj.bar
jsonify_foo = jsonify.jsonify.when("isinstance(obj, Foo)")(jsonify_foo)


def test_list():
    d = ['a', 1, 'b', 2]
    encoded = jsonify.encode(d)
    assert encoded == '["a", 1, "b", 2]'

def test_list_iter():
    d = range(3)
    encoded = jsonify.encode_iter(d)
    assert ''.join(jsonify.encode_iter(d)) == jsonify.encode(d)

def test_dictionary():
    d = {'a': 1, 'b': 2}
    encoded = jsonify.encode(d)
    assert encoded == '{"a": 1, "b": 2}'

def test_specificjson():
    a = Foo("baz")
    encoded = jsonify.encode(a)
    assert encoded == '"foo-baz"'

def test_specific_in_list():
    a = Foo("baz")
    d = [a]
    encoded = jsonify.encode(d)
    assert encoded == '["foo-baz"]'

def test_specific_in_dict():
    a = Foo("baz")
    d = {"a": a}
    encoded = jsonify.encode(d)
    assert encoded == '{"a": "foo-baz"}'

def test_nospecificjson():
    b = Baz()
    try:
        encoded = jsonify.encode(b)
    except Exception, e:
        encoded = e.__class__.__name__
    assert encoded == 'NoApplicableMethods'

def test_exlicitjson():
    b = Bar("bq")
    encoded = jsonify.encode(b)
    assert encoded == '"bar-bq"'

def test_exlicitjson_in_list():
    b = Bar("bq")
    d = [b]
    encoded = jsonify.encode(d)
    assert encoded == '["bar-bq"]'

def test_exlicitjson_in_dict():
    b = Bar("bq")
    d = {"b": b}
    encoded = jsonify.encode(d)
    assert encoded == '{"b": "bar-bq"}'
